/*
 * Copyright (C) 2006-2017,2020 Christian Pierre MOMON
 * 
 * This file is part of Xidyn.
 * 
 * Xidyn is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Xidyn is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Xidyn.  If not, see <http://www.gnu.org/licenses/>
 */
package fr.devinsy.xidyn.data;

import java.io.Serializable;

import fr.devinsy.strings.StringsUtils;

/**
 * The Class SimpleTagData.
 * 
 * IdData class is used to hold application data and the business logic that
 * operates on the data.
 * 
 * The only requirement of a IdData class is that it must implement a display
 * method. The display method must return a text representation of the data,
 * suitable for display in a web page.
 * 
 * XID provides a User Input IdData, Text IdData and ... application may also
 * implement it's own IdData classes.
 * 
 */
public class SimpleTagData implements Serializable, TagData
{
    private static final long serialVersionUID = 8976245034682639923L;

    public enum IterationStrategy
    {
        ONLY_FIRST_ROW,
        ONLY_FIRST_TWO_ROWS,
        ONLY_ROWS_WITH_ID,
        ONLY_ROWS_WITHOUT_ID,
        ALL_ROWS
    }

    IterationStrategy iterationStrategy;
    private TagAttributes attributes;
    private boolean excludeSection;
    private DisplayMode displayMode = DisplayMode.REPLACE;
    private String content;

    /**
     * Instantiates a new simple tag data.
     */
    public SimpleTagData()
    {
        this(null);
    }

    /**
     * Instantiates a new simple tag data.
     * 
     * @param text
     *            the text
     */
    public SimpleTagData(final String text)
    {
        this.attributes = new TagAttributes();
        this.excludeSection = false;
        this.displayMode = DisplayMode.REPLACE;
        this.content = text;
        this.iterationStrategy = IterationStrategy.ALL_ROWS;
    }

    /**
     * Append attribute.
     *
     * @param label
     *            the label
     * @param value
     *            the value
     */
    public void appendAttribute(final String label, final String value)
    {
        this.attributes.appendAttribute(label, value);
    }

    /**
     * Append content.
     * 
     * @param text
     *            the text
     */
    public void appendContent(final String text)
    {
        if (this.content == null)
        {
            this.content = text;
        }
        else
        {
            this.content += text;
        }
    }

    /**
     * Attributes.
     * 
     * @return the tag attributes
     */
    public TagAttributes attributes()
    {
        TagAttributes result;

        result = this.attributes;

        //
        return result;
    }

    /**
     * Display mode.
     * 
     * @return the mode
     */
    public DisplayMode displayMode()
    {
        DisplayMode result;

        result = this.displayMode;

        return result;
    }

    /**
     * Exclude section.
     * 
     * @return true, if successful
     */
    public boolean excludeSection()
    {
        boolean result;

        result = this.excludeSection;

        //
        return result;
    }

    /**
     * Gets the attribute.
     *
     * @param label
     *            the label
     * @return the attribute
     */
    public TagAttribute getAttribute(final String label)
    {
        TagAttribute result;

        result = this.attributes.getAttribute(label);

        //
        return result;
    }

    /**
     * Gets the attribute mode.
     *
     * @param label
     *            the label
     * @return the attribute mode
     */
    public DisplayMode getAttributeMode(final String label)
    {
        DisplayMode result;

        TagAttribute attribute = this.attributes.get(label);
        if (attribute == null)
        {
            result = null;
        }
        else
        {
            result = attribute.getMode();
        }

        //
        return result;
    }

    /**
     * Gets the attribute value.
     *
     * @param label
     *            the label
     * @return the attribute value
     */
    public String getAttributeValue(final String label)
    {
        String result;

        TagAttribute attribute = this.attributes.get(label);
        if (attribute == null)
        {
            result = null;
        }
        else
        {
            result = attribute.getValue();
        }

        //
        return result;
    }

    /**
     * Display.
     * 
     * @return the string
     */
    public String getContent()
    {
        String result;

        result = this.content;

        //
        return result;
    }

    /**
     * Iteration strategy.
     * 
     * @return the iteration strategy
     */
    public IterationStrategy iterationStrategy()
    {
        IterationStrategy result;

        result = this.iterationStrategy;

        //
        return result;
    }

    /**
     * Sets the attribute.
     *
     * @param label
     *            the label
     * @param value
     *            the value
     */
    public void setAttribute(final String label, final String value)
    {
        if (this.attributes.containsKey(label))
        {
            this.attributes.get(label).setValue(value);
        }
        else
        {
            if (StringsUtils.containsAnyIgnoreCase(label, "style", "class"))
            {
                this.attributes.put(label, new TagAttribute(label, value, DisplayMode.APPEND));
            }
            else
            {
                this.attributes.put(label, new TagAttribute(label, value));
            }
        }
    }

    /**
     * Sets the attribute.
     *
     * @param label
     *            the label
     * @param value
     *            the value
     * @param mode
     *            the mode
     */
    public void setAttribute(final String label, final String value, final DisplayMode mode)
    {
        if (this.attributes.containsKey(label))
        {
            TagAttribute attribute = this.attributes.get(label);
            attribute.setValue(value);
            attribute.setMode(mode);
        }
        else
        {
            this.attributes.put(label, new TagAttribute(label, value, mode));
        }
    }

    /**
     * Sets the attribute mode.
     *
     * @param label
     *            the label
     * @param mode
     *            the mode
     */
    public void setAttributeMode(final String label, final DisplayMode mode)
    {
        if (this.attributes.containsKey(label))
        {
            this.attributes.get(label).setMode(mode);
        }
        else
        {
            TagAttribute tag = new TagAttribute(label, null, mode);
            this.attributes.put(label, tag);
        }
    }

    /**
     * Sets the content.
     * 
     * @param text
     *            the new content
     */
    public void setContent(final String text)
    {
        this.content = text;
    }

    /**
     * Sets the display mode.
     * 
     * @param displayMode
     *            the new display mode
     */
    public void setDisplayMode(final DisplayMode displayMode)
    {
        this.displayMode = displayMode;
    }

    /**
     * Sets the exclude section.
     * 
     * @param excludeSection
     *            the new exclude section
     */
    public void setExcludeSection(final boolean excludeSection)
    {
        this.excludeSection = excludeSection;
    }

    /**
     * Sets the iteration strategy.
     * 
     * @param strategy
     *            the new iteration strategy
     */
    public void setIterationStrategy(final IterationStrategy strategy)
    {
        this.iterationStrategy = strategy;
    }
}
